import { ClientOptions, Collection, EmbedBuilder, GuildMember, Message, Snowflake, User } from "discord.js";
import Elo from "arpad";
import { Sequelize } from "sequelize";
import { Battle, BattleState } from "./battle";
import { CHARACTERS } from "./characters";
import { CommandClient } from "./commandclient";
import { Player } from "./models/player";
import { Unit } from "./models/unit";

const ELO = new Elo();

export class Challenge {
    constructor (player1: Player, player2: Player, message: Message) {
        this.player1 = player1;
        this.player2 = player2;
        this.message = message
        this.id = player1.user_id + player2.user_id;
    }
    
    player1: Player
    player2: Player
    message: Message
    id: string
    created = Date.now()
}

export class Trade {
    constructor (origin: User, tradeguy: User, unit1: Unit, unit2: Unit, message: Message) {
        this.origin = origin;
        this.tradeguy = tradeguy;
        this.unit1 = unit1;
        this.unit2 = unit2;
        this.message = message
        this.id = 't' + unit1.user_id + unit2.user_id;
    }
    
    origin: User
    tradeguy: User
    unit1: Unit
    unit2: Unit
    message: Message
    id: string
    created = Date.now()
}

export class ActiveBattle {
    constructor (battle: Battle, message: Message) {
        this.battle = battle;
        this.message = message;
    }

    async sendRound () {
        try {
            this.rounds++;
            const outcome = this.battle.simulateRound();
            const log = this.battle.log.slice(0, 1024);

            if (this.battle.channel.isDMBased()) return;
            const player1 = this.battle.channel.guild.members.resolve(this.battle.player1.user_id);
            const player2 = this.battle.channel.guild.members.resolve(this.battle.player2.user_id);
            
            await this.message.edit({
                embeds: [
                    new EmbedBuilder()
                        .addFields([
                            { name: player1?.user.username || '████████', value: `${this.battle.team1[0].character.name} [${this.battle.team1[0].health}]\n${this.battle.team1[1].character.name} [${this.battle.team1[1].health}]\n${this.battle.team1[2].character.name} [${this.battle.team1[2].health}]`, inline: true },
                            { name: player2?.user.username || '████████', value: `${this.battle.team2[0].character.name} [${this.battle.team2[0].health}]\n${this.battle.team2[1].character.name} [${this.battle.team2[1].health}]\n${this.battle.team2[2].character.name} [${this.battle.team2[2].health}]`, inline: true },
                            { name: 'Log', value: log }
                        ])
                        .setColor(0x8c110b)
                        .setThumbnail('https://i.imgur.com/sMrWQWO.png')
                ]
            }).catch(this.logAndSelfDestruct);

            if (outcome === BattleState.Ongoing && this.rounds < 20) {this.timeout = setTimeout(this.sendRound.bind(this), 4_000);return;}

            this.battle.client.ACTIVE_BATTLES.splice(this.battle.client.ACTIVE_BATTLES.indexOf(this), 1);
            
            let elo1 = this.battle.player1.elo;
            let elo2 = this.battle.player2.elo;
            
            switch (outcome) {
                case BattleState.Team1Win:
                    await this.battle.player1.update({ elo: ELO.newRatingIfWon(elo1, elo2) });
                    await this.battle.player2.update({ elo: ELO.newRatingIfLost(elo2, elo1) });

                    await this.battle.channel.send({
                        content: `${player1 || '████████'} has defeated ${player2 || '████████'}!\n\n${player1?.user.username || '████████'} +${this.battle.player1.elo-elo1}\n${player2?.user.username || '████████'} -${elo2-this.battle.player2.elo}`
                    }).catch(this.logAndSelfDestruct);
                    break;
                case BattleState.Team2Win:
                    await this.battle.player2.update({ elo: ELO.newRatingIfWon(elo2, elo1) });
                    await this.battle.player1.update({ elo: ELO.newRatingIfLost(elo1, elo2) });

                    await this.battle.channel.send({
                        content: `${player2 || '████████'} has defeated ${player1 || '████████'}!\n\n${player2?.user.username || '████████'} +${this.battle.player2.elo-elo2}\n${player1?.user.username || '████████'} -${elo1-this.battle.player1.elo}`
                    }).catch(this.logAndSelfDestruct);
                    break;
            }
        }
        catch (err) {
            this.logAndSelfDestruct(err);
        }
    }

    static async create (battle: Battle) {
        const activeBattle = new ActiveBattle(battle, (await battle.channel.send('initial battle state')));
        return activeBattle;
    }

    logAndSelfDestruct (err: unknown) {
        console.error(err);
        this.battle.client.ACTIVE_BATTLES.splice(this.battle.client.ACTIVE_BATTLES.indexOf(this), 1);
    }

    rounds = 0
    battle: Battle
    message: Message
    timeout = setTimeout(this.sendRound.bind(this), 2_000)
}

export class CBClient extends CommandClient {
    constructor (options: ClientOptions, db: Sequelize) {
        super(options, db);

        this.on('interactionCreate', async (int) => {
            if (!int.isButton() || !int.channel || int.channel.isDMBased()) return;

            if (int.customId.startsWith('t')){
                const trade = this.TRADES.get(int.customId);
                if (!trade || trade.created+60_000 < Date.now()) {
                    int.reply({
                        content: "That trade has expired!",
                        ephemeral: true
                    });
                    return;
                }

                if (int.user.id != trade.tradeguy.id) {
                    int.reply({
                        content: "You can't accept someone else's trade!",
                        ephemeral: true
                    });
                    return;
                }

                const player1 = await Player.findOne({ where: { user_id: trade.unit1.user_id } });
                const player2 = await Player.findOne({ where: { user_id: trade.unit2.user_id } });

                if (!player1 || !player2) return;

                if (player1.slot1 === trade.unit1.id) player1.set({slot1:null});
                if (player1.slot2 === trade.unit1.id) player1.set({slot2:null});
                if (player1.slot3 === trade.unit1.id) player1.set({slot3:null});
                if (player2.slot1 === trade.unit2.id) player2.set({slot1:null});
                if (player2.slot2 === trade.unit2.id) player2.set({slot2:null});
                if (player2.slot3 === trade.unit2.id) player2.set({slot3:null});

                await trade.unit1.update({ user_id: trade.tradeguy.id });
                await trade.unit2.update({ user_id: trade.origin.id });
                await player1.save();
                await player2.save();

                this.TRADES.delete(trade.id);

                int.reply('Trade accepted!');
                return;
            }
            else {
                const defaultUnit = await Unit.findOne({ where: { id: 0 } }) as Unit;

                const challenge = this.CHALLENGES.get(int.customId);
                if (!challenge || challenge.created+60_000 < Date.now()) {
                    int.reply({
                        content: "That challenge has expired!",
                        ephemeral: true
                    });
                    return;
                }

                if (int.user.id != challenge.player2.user_id) {
                    int.reply({
                        content: "You can't accept someone else's challenge!",
                        ephemeral: true
                    });
                    return;
                }

                if (this.ACTIVE_BATTLES.find(e => {
                    return [challenge.player1.user_id, challenge.player2.user_id].includes(e.battle.player1.user_id)
                    || [challenge.player1.user_id, challenge.player2.user_id].includes(e.battle.player2.user_id);
                })) {
                    int.reply({
                        content: "A participant is already in a battle!",
                        ephemeral: true
                    });
                    return;
                }

                await challenge.message.delete();
                this.CHALLENGES.delete(challenge.id);

                const team1: [Unit, Unit, Unit] = [
                    await Unit.findOne({ where: { id: challenge.player1.slot1 || -1 } }) || defaultUnit,
                    await Unit.findOne({ where: { id: challenge.player1.slot2 || -1 } }) || defaultUnit,
                    await Unit.findOne({ where: { id: challenge.player1.slot3 || -1 } }) || defaultUnit
                ]
                const team2: [Unit, Unit, Unit] = [
                    await Unit.findOne({ where: { id: challenge.player2.slot1 || -1 } }) || defaultUnit,
                    await Unit.findOne({ where: { id: challenge.player2.slot2 || -1 } }) || defaultUnit,
                    await Unit.findOne({ where: { id: challenge.player2.slot3 || -1 } }) || defaultUnit
                ]

                if (challenge.message.channel.isDMBased()) return;
                const player1 = challenge.message.channel.guild.members.resolve(challenge.player1.user_id);
                const player2 = challenge.message.channel.guild.members.resolve(challenge.player2.user_id);
                
                const battle = new Battle(this, int.channel, challenge.player1, challenge.player2, team1, team2);
                const activeBattle = new ActiveBattle(battle,
                    await int.channel.send({
                        embeds: [
                            new EmbedBuilder()
                                .addFields([
                                    { name: player1?.user.username || '████████', value: `${battle.team1[0].character.name} [${battle.team1[0].health}]\n${battle.team1[1].character.name} [${battle.team1[1].health}]\n${battle.team1[2].character.name} [${battle.team1[2].health}]`, inline: true },
                                    { name: player2?.user.username || '████████', value: `${battle.team2[0].character.name} [${battle.team2[0].health}]\n${battle.team2[1].character.name} [${battle.team2[1].health}]\n${battle.team2[2].character.name} [${battle.team2[2].health}]`, inline: true },
                                    { name: 'Log', value: 'Starting...' }
                                ])
                                .setColor(0x8c110b)
                                .setThumbnail('https://i.imgur.com/sMrWQWO.png')
                        ]
                    }
                ));
                this.ACTIVE_BATTLES.push(activeBattle);
            }
        });
    }

    readonly CHALLENGES: Collection<string, Challenge> = new Collection()
    readonly TRADES: Collection<string, Trade> = new Collection()
    readonly ACTIVE_BATTLES: ActiveBattle[] = []

    static async createPlayer(user_id: Snowflake) {
        return Player.create({
            user_id: user_id,
            start: 1680323712
        });
    }

    static async initializePlayer(user_id: Snowflake) {
        Unit.create({ user_id: user_id, character_id: CHARACTERS.randomKey()! });
        Unit.create({ user_id: user_id, character_id: CHARACTERS.randomKey()! });
        Unit.create({ user_id: user_id, character_id: CHARACTERS.randomKey()! });
        return CBClient.createPlayer(user_id);
    }

    static async findOrCreatePlayer(user_id: Snowflake) {
        const player = await Player.findOne({
            where: {
                user_id: user_id
            }
        });

        if (!player) return await CBClient.initializePlayer(user_id);
        return player;
    }

    static async spendOnPlayer(player: Player, amount: number) {
        const currency = Date.now()/1000 - player.start - player.spent;
        console.debug(Date.now()/1000, player.start, player.spent, currency);
        
        if (currency - amount < 0) return false;
        await player.update({ spent: player.spent+amount });
        return true;
    }

    static async grantMoneyToPlayer(player: Player, amount: number) {
        await player.update({ spent: player.spent-amount });
        return;
    }
}